/*
 * Copyright (C) 2004 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*
 * $Log$
 */

#include <dev/debug.h>
#ifdef ETHERNUT2
#include <dev/lanc111.h>
#else
#include <dev/nicrtl.h>
#endif
#include <sys/thread.h>
#include <sys/timer.h>
#include <sys/version.h>
#include <sys/confnet.h>
#include <sys/confos.h>

#include <sys/confnet.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <netinet/in.h>
#include <pro/dhcp.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <io.h>

#include "discover.h"

static char *version = "1.0.1";

/*
 * Device used for stdout. 
 */
#define STDOUT_DEV      devDebug0

/*
 * Symbolic name of the stdout device.
 */
#define STDOUT_DEVNAME  "uart0"

/*
 * Bitrate of the stdout device.
 */
#define STDOUT_BAUDRATE 115200

/*
 * Default MAC address.
 *
 * Used if none has been found in EEPROM.
 */
#define STDNET_MAC      { 0x00, 0x06, 0x98, 0x20, 0x00, 0x00 }

/*
 * Default IP address and network mask.
 *
 * Used if not found in EEPROM and if no DHCP response received.
 */
#define STDNET_IPADDR   "192.168.192.202"
#define STDNET_IPMASK   "255.255.255.0"


/*
 * Open a standard output device.
 */
int InitializeConsole(void)
{
    u_long baud = 115200;

    if(NutRegisterDevice(&devDebug0, 0, 0)) {
        return -1;
    }

    if(freopen("uart0", "w", stdout) == 0) {
        return -1;
    }
    return _ioctl(_fileno(stdout), UART_SETSPEED, &baud);
}

/*
 * Open a network device.
 */
int InitializeNetwork(void)
{
    u_char null_mac[] = STDNET_MAC;
    int rc;

    rc = NutRegisterDevice(&DEV_ETHER, 0x8300, 5);

    if(rc == 0) {
        /* Try DHCP with MAC from EEPROM. */
        rc = NutDhcpIfConfig("eth0", 0, 60000);
        if (rc) {
            /* No valid EEPROM contents, use hard coded MAC. */
            rc = NutDhcpIfConfig("eth0", null_mac, 60000);
            if (rc) {
                /* No DHCP server found, use hard coded IP. */
                u_long ip_addr = inet_addr(STDNET_IPADDR);
                u_long ip_mask = inet_addr(STDNET_IPMASK);

                rc = NutNetIfConfig("eth0", null_mac, ip_addr, ip_mask);
            }
        }
    }
    return rc;
}

/*
 * Broadcast an announcement datagram.
 */
void Announce(UDPSOCKET * sock, u_long xid, DISCOVER *tele)
{
    memset(tele, 0, sizeof(DISCOVER));

    tele->tele_type = 1;
    tele->tele_xid = xid;
    memcpy(tele->tele_hostname, confos.hostname, sizeof(tele->tele_hostname) - 1);
    memcpy(tele->tele_mac, confnet.cdn_mac, sizeof(tele->tele_mac));
    tele->tele_ip_addr = confnet.cdn_ip_addr;
    tele->tele_ip_mask = confnet.cdn_ip_mask;
    tele->tele_gateway = confnet.cdn_gateway;
    tele->tele_cip_addr = confnet.cdn_cip_addr;

    NutUdpSendTo(sock, INADDR_BROADCAST, DISCOVER_PORT, tele, sizeof(DISCOVER));
}

/*
 * Discover service thread.
 */
THREAD(DiscoverServer, arg)
{
    UDPSOCKET *sock;
    DISCOVER *tele;
    u_long raddr;
    u_short rport;
    int num;
    u_long xid;

    /* Create a hopefully unique identifier. */
    xid = NutGetTickCount();
    for(num = sizeof(confnet.cdn_mac); --num; xid <<= 4) {
        xid += confnet.cdn_mac[num];
    }

    /* Send an initial announce telegram. */
    sock = NutUdpCreateSocket(DISCOVER_PORT);
    tele = malloc(sizeof(DISCOVER));
    Announce(sock, xid, tele);

    /* 
     * Endlessly loop for incoming discover telegrams. 
     */
    for (;;) {
        num = NutUdpReceiveFrom(sock, &raddr, &rport, tele, sizeof(DISCOVER), DISCOVER_RATE);
        if (num == 0) {
            /* Timeout, send an announcement. */
            Announce(sock, xid, tele);
        }
        else if (num == sizeof(DISCOVER)) {
            /* Telegram type 0, request from host. */
            if (tele->tele_type == 0) {
                Announce(sock, ++xid, tele);
            }
            /* Telegram type 2, update from host. ID must match. */
            else if (tele->tele_type == 2 && tele->tele_xid == xid) {
                /* Save updated host name. */   
                memcpy(confos.hostname, tele->tele_hostname, sizeof(confos.hostname) - 1);
                NutSaveConfig();

                /* Save updated network configuration. */
                memcpy(confnet.cdn_mac, tele->tele_mac, sizeof(confnet.cdn_mac));
                confnet.cdn_ip_addr = tele->tele_ip_addr;
                confnet.cdn_ip_mask = tele->tele_ip_mask;
                confnet.cdn_gateway = tele->tele_gateway;
                confnet.cdn_cip_addr = tele->tele_cip_addr;
                NutNetSaveConfig();

                /* Send announcement with new data. */
                Announce(sock, ++xid, tele);
            }
        }

        /* Avoid responding to crazy hosts too often. */
        NutSleep(2000);
    }
}

/*
 * Main application routine. 
 */
int main(void)
{
    int err_con; /* 0, if console interface OK. */
    int err_net; /* 0, if network interface OK. */

    /*
     * Initialize standard output interface.
     * If successfull, display some information.
     */
    err_con = InitializeConsole();
    if(err_con == 0) {
        printf("\n\nDiscovery Service Version %s\n", version);
        printf("Nut/OS %s\n", NutVersionString());
        printf("Compiled by ");
#ifdef __IMAGECRAFT__
        printf("ICCAVR");
#else
        printf("GCC");
#endif
        printf(" for ");
#ifdef __AVR_ATmega128__
        puts("ATmega128");
#else
        puts("ATmega103");
#endif
    }

    /*
     * Initialize network interface.
     */
    err_net = InitializeNetwork();
    if(err_con == 0) {
        if(err_net) {
            puts("Error: No network");
        } else {
            printf("Listening at %s:%u\n", inet_ntoa(confnet.cdn_ip_addr), DISCOVER_PORT);
        }
    }

    /*
     * Start discover server.
     */
    if(err_net == 0) {
        NutThreadCreate("discd", DiscoverServer, 0, 768);
    }

    /*
     * We could do something more useful here.
     */
    for (;;) {
        NutSleep(30000);
        if(err_con == 0) {
            puts("It's so boring...");
        }
    }
}

